package mthsc865.datastructures;

import java.util.Collection;
import java.util.Iterator;
import java.util.ConcurrentModificationException;

/**
 * This class implements a Stack using a simple Linked List.
 *
 * @author Daniel D. Warner
 * @version 2.0 09/07/04
 */
public class StackLL extends AbstractStack {
    /**
     * The Linked List starts with this link.
     */
    protected Link top;
    
    /**
     * The Linked List is built from a simple forward link.
     */
    protected class Link {
        /**
         * Each link contains a reference to the next link and
         * a reference to the object stored in the link.
         */
        Link next;
        Object data;
        Link(Link lnk, Object obj) {
            next = lnk;
            data = obj;
        }
        
        Link(Object obj) {
            next = null;
            data = obj;
        }
        
        Link() {
            next = null;
            data = null;
        }
    }
    
    /**
     * Constructs a stack which uses a simple Linked List.
     */
    public StackLL() {
        hash = 0;
        top = new Link( new Integer(0));
    }
    
    /**
     * Constructs a stack which is initialized with the objects
     * in the specified collection
     * @param c the collection of objects to be initially pushed
     * onto this stack.
     */
    public StackLL(Collection c) {
        hash = 0;
        top = new Link( new Integer(0));
        Iterator ci = c.iterator();
        while (ci.hasNext()) {
            this.push( ci.next() );
        }
    }
    
    /**
     * Constructs a stack with a specified hash code and a specified
     * linked list.  This constuctor is only for use within the class
     * and its subclasses.
     * @param hash the hash code
     * @param topOfLinkedList the top of the linked list.
     */
     protected StackLL(int hash, Link topOfLinkedList) {
        this.hash = hash;
        top = topOfLinkedList;
    }
    
    /**
     * Pushes an object onto the top of the stack.
     * @param obj the object to be placed on the top of the stack.
     */
    public void push(Object obj) {
        updateHashCode(obj);
        Link lnk = new Link(top.next, obj);
        top.next = lnk;
        top.data = new Integer(1+((Integer)top.data).intValue());
    }
    
    /**
     * Removes and returns the object on the top of the stack.
     * @return the object on the top of the stack or null if the stack is empty.
     */
    public Object pop() {
        if (top.next == null) {
            return null;
        } else {
            Object obj = top.next.data;
            top.next = top.next.next;
            top.data = new Integer(((Integer)top.data).intValue()-1);
            downdateHashCode(obj);
            return obj;
        }
    }
    
    /**
     * Returns the object on the top of the stack.
     * The stack remains unchanged.
     * @return the object on the top of the stack or null if the stack is empty.
     */
    public Object peek() {
        if (top.next == null) {
            return null;
        } else {
            return top.next.data;
        }
    }
    
    /**
     * Removes all the objects from this stack.
     * This is one of the optional operations specified by the Collection Interface
     */
    public void clear() {
        hash = 0;
        Link thisLink = top;
        Link nextLink = top.next;
        top.next = null;
        top.data = new Integer(0);
        // The following loop is included to make life easier for garbage collection
        while (nextLink != null) {
            thisLink = nextLink;
            nextLink = nextLink.next;
            thisLink.next = null;
        }
    }
    
    /**
     * The Cloneable Interface indicates that the clone
     * method, which is inherited from Object, is implemented
     * so that a true clone (a true and independent copy) of 
     * the object is returned.
     * @return a clone (a true and independent copy) of this stack.
     */
    public Object clone() {
        int mySize = ((Integer)top.data).intValue();
        Link newTop = new Link(new Integer(mySize));
        Iterator iter = this.iterator();
        Link oldLink = newTop;
        Link newLink;
        while (iter.hasNext()) {
            newLink = new Link(iter.next());
            oldLink.next = newLink;
            oldLink = newLink;
        }
        return (new StackLL(hash, newTop));
    }
    
    /**
     * Determines whether the stack is empty.
     * @return true - if this collection contains no elements.
     */
    public boolean isEmpty() {
        // The isEmpty() method is being overidden in order to check the
        // assertion that: top.next == null  if and only if  top.data is zero.
        boolean b1,  b2;
        int mySize = ((Integer)top.data).intValue();
        b1 = (mySize == 0);
        b2 = (top.next == null);
        if ((b1 && !b2) || (!b1 && b2)) {
            System.out.println("*** StackLL isEmpty assertion error ***");
        }
        return b1;
    }
    
    /**
     * An iterator for an ArrayList stack.
     */
    public class StackLLIterator implements Iterator {
        private StackLL myS;  // a reference to my Stack
        private Link index;   // my current position in the stack
        private long check;   // my copy of the initial hash code. It will be an
                              // error if the stack is altered during the iteration.
        /**
         * Constructor for a StackLLIterator
         * @param s the stack for this iterator.
         */
        protected StackLLIterator(StackLL s) {
            myS = s;
            index = myS.top.next;
            check = myS.hash;
        }
        
        /**
         * Determines if there is an object in stack that can be returned by next().
         * @return true - if there is an object in the stack that can be returned by next().
         */
        public boolean hasNext() {
            //  Check that the stack has not been altered during the processing
            //  of this iterator
            if (check == myS.hash) {
                return (index != null);
            } else {
                throw( new ConcurrentModificationException("This Stack has been altered"));
            }
        }
        
        /**
         * Returns the next object in stack, if there is one.
         * It throws a ConcurrentModificationException if the stack has been changed
         * during the iteration.
         * @return the next object in the stack.
         */
        public Object next() {
            if (check == myS.hash) {
                Object obj = index.data;
                index = index.next;
                return obj;
            } else {
                throw( new ConcurrentModificationException("This Stack has been altered"));
            }
        }
	
        /**
         * This method is an optional method of the Iterator interface. 
         * If implemented, it would remove from the stack the last element returned
         * by the iterator. Since this would violate the integrity of the abstract stack 
         * data structure it is not implemented.
         * Throws an UnsupportedOperationException.
         */
        public void remove() {
            throw( new UnsupportedOperationException(
            "remove is not implemented for mthsc865.datastructure.StackAL"));
        }
    }
    
    /**
     * A factory method which returns an Iterator to the collection in this stack.
     * This iterator will walk throught the stack from the top to the bottom.
     * It will throw a ConcurrentModificationException if the stack is altered during
     * the iteration.
     * @return a StackAL iterator.
     */
    public Iterator iterator() {
        return new StackLLIterator(this);
    }
     
    /**
     * Determines the size of this stack.
     * @return the number of objects in this stack. 
     */
    public int size() {
        return ((Integer)top.data).intValue();
    }
    
    /**
     * List the objects in the stack
     * @return a formatted string listing the objects in this stack
     */
    public String toString() {
        String s = super.toString();
        return( "StackLL" + s.substring(5, s.length()) );
    }
    
    /**
     * This main method tests the StackLL class to insure that the
     * elementary functions are correct.
     * @param args optional command line arguments which will be ignored.
     */
    public static void main(String[] args) {
        StackLL s1 = new StackLL();
        String name;
        String linesep = System.getProperty("line.separator");
        
        // Place Alice, Bob, and Carol on the stack and print out the
        // stack and its basic properties.
        s1.push("Alice");
        s1.add("Bob");
        s1.push("Carol");
        System.out.println(s1);
        System.out.println("s1.hashCode: " + s1.hashCode());
        System.out.println("Empty? " + s1.isEmpty());
        
        // Test the iterator
        System.out.println(linesep + "s1 Iterator test:");
        Iterator iter = s1.iterator();
        while (iter.hasNext()) {
            name = (String)iter.next();
            System.out.println(name);
        }
        
        // Create a second and third stack
        System.out.println(linesep + "Looking at three stacks");
        StackLL s2 = new StackLL(s1);
        System.out.println(linesep + "s2 Iterator test:");
        iter = s2.iterator();
        while (iter.hasNext()) {
            name = (String)iter.next();
            System.out.println(name);
        }
        StackLL s3 = new StackLL();
        s3.addAll(s2);
        System.out.println(linesep + "s3 Iterator test:");
        iter = s3.iterator();
        while (iter.hasNext()) {
            name = (String)iter.next();
            System.out.println(name);
        }
        
        // Manipulate the stacks and check the different methods
        // particularly the hash code, equality, and contains.
        System.out.println("s1 equals s2? " + s1.equals(s2));
        System.out.println("s1 equals s3? " + s1.equals(s3));
        System.out.println("s1.hashCode: " + s1.hashCode());
        System.out.println("s2.hashCode: " + s2.hashCode());
        System.out.println("s3.hashCode: " + s3.hashCode());
        System.out.println("Is Bob in s1? " + s1.contains("Bob"));
        s1.add("Doug");
        System.out.println(s1.peek());
        System.out.println("s1.hashCode: " + s1.hashCode());
        System.out.println("s2.hashCode: " + s2.hashCode());
        System.out.println("s3.hashCode: " + s3.hashCode());
        System.out.println("s1 equals s2? " + s1.equals(s2));
        System.out.println("s1 equals s3? " + s1.equals(s3));
        System.out.println("Is s1 contained in s2? " + s2.containsAll(s1));
        System.out.println("Is s2 contained in s1? " + s1.containsAll(s2));
        System.out.println("Is s2 contained in s3? " + s3.containsAll(s3));
        
        // Check out the cloning mechanism
        System.out.println("\nLooking at cloning");
        StackLL s4 = (StackLL)s1.clone();
        System.out.println("s1 equals s4? " + s1.equals(s4));
        System.out.println("s1.hashCode: " + s1.hashCode());
        System.out.println("s4.hashCode: " + s4.hashCode());
        System.out.println("s4.pop: " + s4.pop());
        System.out.println("s1 equals s4? " + s1.equals(s4));
        System.out.println("s1.hashCode: " + s1.hashCode());
        System.out.println("s4.hashCode: " + s4.hashCode());
        
        // Check out the conversion from a stack to an array.
        System.out.println("\nLooking at an array and a stack");
        Object[] names = s1.toArray();
        for (int i=0; i<names.length; i++) {
            System.out.println(names[i]);
        }
        String[] moreNames = (String[])s1.toArray(new String[5]);
        System.out.println("moreNames.length: " + moreNames.length);
        for (int i=0; i<moreNames.length; i++) {
            System.out.println(moreNames[i]);
        }
        
        // Empty the stacks
        System.out.println(linesep + "Emptying the stacks");
        System.out.println(s1.pop());
        System.out.println(s1.hashCode());
        System.out.println(s1.pop());
        System.out.println(s1.pop());
        System.out.println(s1.pop());
        System.out.println("Empty? " + s1.isEmpty());
        s2.clear();
        System.out.println("Empty? " + s2.isEmpty());
        System.out.println("s1 equals s2? " + s1.equals(s2));
        System.out.println("s1.hashCode: " + s1.hashCode());
        System.out.println("s2.hashCode: " + s2.hashCode());
        
        // Test the iterator for consistency
        System.out.println(linesep + "Testing the iterator for consistency checking");
        iter = s4.iterator();
        try {
            System.out.println(iter.next());
            System.out.println(iter.next());
            System.out.println("Pop: " + s4.pop());
            System.out.println(iter.next());
        } catch (ConcurrentModificationException e) {
            System.out.println("Concurrent Modification Exception: " + e.getMessage());
        }
    }
}